/*
                        Compile
    Linux:
        gcc ./fwupd.c -I /usr/include/PCSC -l pcsclite -o fwupd

    macOS:
        gcc ./fwupd.c -framework PCSC -o fwupd

    macOS Sonoma driver workaround:
        sudo defaults write /Library/Preferences/com.apple.security.smartcard useIFDCCID -bool yes

    For Windows x86_64
        x86_64-w64-mingw32-gcc ./fwupd.c -l winscard -o fwupd.exe

    For Windows i686
        i686-w64-mingw32-gcc ./fwupd.c -l winscard -o fwupd.exe
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef __MINGW32__
#include <winscard.h>
#else
#include <PCSC/winscard.h>
#include <PCSC/wintypes.h>
#define SCARD_ATTR_ATR_STRING 0x00090303
#endif

#define EUICC_INTERFACE_BUFSZ 264

#define FWUPD_BLOCK_SIZE 2061
#define FWUPD_TRANS_BLOCK_MIN_SIZE 6
#define FWUPD_TRANS_BLOCK_MAX_SIZE 255

#define APDU_FWUPD_SELECT_AID "\x01\xA4\x04\x00\x10\xA0\x65\x73\x74\x6B\x6D\x65\xFF\xFF\xFF\xFF\x66\x77\x75\x70\x64"
#define APDU_FWUPD_TRIGGER "\x01\x55\x55\x55\x00"
#define APDU_GET_VERSION "\xAA\xFF\x00\x00\x08"

#define ATR_FPR_FWUPD "estk.me fwupd"
#define ATR_FPR_NOT_FWUPD "estk.me"

#ifdef __MINGW32__
#define LINE_BREAK "\r\n"
#else
#define LINE_BREAK "\n"
#endif

// clang-format off
#define fprintlnf(f, format, ...) fprintf(f, format LINE_BREAK, ## __VA_ARGS__); fflush(f)
#define printlnf(format, ...) fprintlnf(stdout, format, ## __VA_ARGS__)
// clang-format on

#ifdef __MINGW32__
void *memmem(const void *haystack, size_t haystack_len, const void *const needle, const size_t needle_len) {
    if (haystack == NULL || haystack_len == 0 || needle == NULL || needle_len == 0)
        return NULL;
    for (const char *h = haystack; haystack_len >= needle_len; ++h, --haystack_len)
        if (!memcmp(h, needle, needle_len))
            return h;
    return NULL;
}
#endif

static SCARDCONTEXT pcsc_ctx = 0;
static SCARDHANDLE pcsc_hCard = 0;
static LPSTR pcsc_mszReaders = NULL;

typedef struct {
    const uint8_t unlock[5];
    const uint8_t block_write[5];
    const uint8_t block_read[5];
    const uint8_t block_flash[5];
    const uint8_t exit[5];
} context_t;

static context_t context_v1 = {
    .unlock = {0x00, 0x21, 0x00, 0x00, 0x00},
    .block_write = {0x00, 0x11, 0x00, 0x00, 0x00},
    .block_read = {0x00, 0x12, 0x00, 0x00, 0x00},
    .block_flash = {0x00, 0x13, 0x00, 0x00, 0x00},
    .exit = {0x00, 0x00, 0x00, 0x00, 0x00},
};

static context_t context_v2 = {
    .unlock = {0xAA, 0x21, 0x00, 0x00, 0x00},
    .block_write = {0xAA, 0x11, 0x00, 0x00, 0x00},
    .block_read = {0xAA, 0x12, 0x00, 0x00, 0x00},
    .block_flash = {0xAA, 0x13, 0x00, 0x00, 0x00},
    .exit = {0xAA, 0x00, 0x00, 0x00, 0x00},
};

static uint8_t get_mss() {
    const char *value = getenv("FWUPD_MSS");
    if (value == NULL) {
        return FWUPD_TRANS_BLOCK_MAX_SIZE;
    }
    const long long mss = strtol(value, NULL, 10);
    if (mss < FWUPD_TRANS_BLOCK_MIN_SIZE) {
        return FWUPD_TRANS_BLOCK_MIN_SIZE;
    }
    if (mss > FWUPD_TRANS_BLOCK_MAX_SIZE) {
        return FWUPD_TRANS_BLOCK_MAX_SIZE;
    }
    return (uint8_t) mss;
}

static void pcsc_error(const char *method, const int32_t code) {
#ifdef __MINGW32__
    fprintlnf(stderr, "%s failed: %08X", method, code);
#else
    fprintlnf(stderr, "%s failed: %08X (%s)", method, code, pcsc_stringify_error(code));
#endif
}

static int pcsc_ctx_open(void) {
    DWORD dwReaders;

    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;

    int ret = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &pcsc_ctx);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardEstablishContext()", ret);
        return -1;
    }

    ret = SCardListReaders(pcsc_ctx, NULL, NULL, &dwReaders);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardListReaders()", ret);
        return -1;
    }

    pcsc_mszReaders = malloc(sizeof(char) * dwReaders);
    if (pcsc_mszReaders == NULL) {
        fprintlnf(stderr, "malloc: not enough memory");
        return -1;
    }

    ret = SCardListReaders(pcsc_ctx, NULL, pcsc_mszReaders, &dwReaders);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardListReaders()", ret);
        return -1;
    }

    return 0;
}

static int pcsc_iter_reader(int (*callback)(int index, const char *reader)) {
    LPSTR psReader = pcsc_mszReaders;
    for (int i = 0, n = 0;; i++) {
        char *p = pcsc_mszReaders + i;
        if (*p == '\0') {
            const int ret = callback(n, psReader);
            if (ret < 0)
                return -1;
            if (ret > 0)
                return 0;
            if (*(p + 1) == '\0')
                break;
            psReader = p + 1;
            n++;
        }
    }
    return -1;
}

static int pcsc_open_hCard_iter(const int index, const char *reader) {
    DWORD dwActiveProtocol;

    const char *value = getenv("DRIVER_IFID");
    if (value != NULL && strtol(value, NULL, 10) != index) {
        return 0;
    }

    value = getenv("DRIVER_NAME");
    if (value != NULL && strstr(reader, value) == NULL) {
        return 0;
    }

    const int ret = SCardConnect(pcsc_ctx, reader, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T0, &pcsc_hCard, &dwActiveProtocol);
    if (ret == SCARD_S_SUCCESS)
        return 1;
    pcsc_error("SCardConnect()", ret);
    if (ret == SCARD_E_SHARING_VIOLATION)
        return 0;
    return -1;
}

static int pcsc_open_hCard(void) { return pcsc_iter_reader(pcsc_open_hCard_iter); }

static void pcsc_close(void) {
    if (pcsc_mszReaders) {
        free(pcsc_mszReaders);
    }

    if (pcsc_hCard) {
        SCardDisconnect(pcsc_hCard, SCARD_UNPOWER_CARD);
    }

    if (pcsc_ctx) {
        SCardReleaseContext(pcsc_ctx);
    }

    pcsc_ctx = 0;
    pcsc_hCard = 0;
    pcsc_mszReaders = NULL;
}

static int pcsc_transmit_lowlevel(uint8_t *rx, uint32_t *rx_len, const uint8_t *tx, const uint32_t tx_len) {
    DWORD rx_len_merged;

    rx_len_merged = *rx_len;
    const int ret = SCardTransmit(pcsc_hCard, SCARD_PCI_T0, tx, tx_len, NULL, rx, &rx_len_merged);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardTransmit()", ret);
        return -1;
    }

    *rx_len = rx_len_merged;
    return 0;
}

static int card_poweron(void) {
    if (pcsc_ctx_open() < 0) {
        return -1;
    }
    if (pcsc_open_hCard() < 0) {
        return -1;
    }
    return 0;
}

static void card_poweroff(void) { pcsc_close(); }

static void fwupd_enter(void) {
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *) APDU_FWUPD_SELECT_AID, sizeof(APDU_FWUPD_SELECT_AID) - 1);

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *) APDU_FWUPD_TRIGGER, sizeof(APDU_FWUPD_TRIGGER) - 1);
}

static int fwupd_init_seq(void) {
    card_poweron();
    card_poweroff();
    if (card_poweron() < 0) {
        return -1;
    }
    fwupd_enter();
    card_poweroff();
    if (card_poweron() < 0) {
        return -1;
    }
    return 0;
}

static int fwupd_unlock_part(const context_t *ctx) {
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, ctx->unlock, sizeof(ctx->unlock));
    if (pcsc_rx[0] != 0x90) {
        return -1;
    }

    return 0;
}

static int fwupd_block_write(const context_t *ctx, const uint8_t *data, const uint8_t mss) {
    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    uint32_t block_size = FWUPD_BLOCK_SIZE;

    while (block_size) {
        uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
        uint8_t tx_size = block_size > mss ? mss : block_size;
        const uint16_t tx_offset = FWUPD_BLOCK_SIZE - block_size;

        memcpy(pcsc_tx, ctx->block_write, sizeof(ctx->block_write));
        memcpy(pcsc_tx + sizeof(ctx->block_write), data + tx_offset, tx_size);

        pcsc_tx[2] = tx_offset >> 8;
        pcsc_tx[3] = tx_offset & 0xFF;
        pcsc_tx[4] = tx_size;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, tx_size + sizeof(ctx->block_write));
        if (pcsc_rx[0] != 0x90) {
            return -1;
        }

        block_size -= tx_size;
    }

    return 0;
}

static int fwupd_block_verify(const context_t *ctx, const uint8_t *data, const uint8_t mss) {
    uint8_t pcsc_tx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    uint8_t block_buffer[FWUPD_BLOCK_SIZE];
    uint32_t block_size = FWUPD_BLOCK_SIZE;

    while (block_size) {
        uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
        uint8_t rx_size = block_size > mss ? mss : block_size;
        const uint16_t rx_offset = FWUPD_BLOCK_SIZE - block_size;

        memcpy(pcsc_tx, ctx->block_read, sizeof(ctx->block_read));
        pcsc_tx[2] = rx_offset >> 8;
        pcsc_tx[3] = rx_offset & 0xFF;
        pcsc_tx[4] = rx_size;

        pcsc_rx_len = sizeof(pcsc_rx);
        memset(pcsc_rx, 0, sizeof(pcsc_rx));
        pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, pcsc_tx, sizeof(ctx->block_read));
        if (pcsc_rx_len != rx_size + 2) {
            return -1;
        }
        if (pcsc_rx[rx_size] != 0x90) {
            return -1;
        }

        memcpy(block_buffer + rx_offset, pcsc_rx, rx_size);

        block_size -= rx_size;
    }

    if (memcmp(block_buffer, data, FWUPD_BLOCK_SIZE) != 0) {
        return -1;
    }

    return 0;
}

static int fwupd_block_flash(const context_t *ctx) {
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, ctx->block_flash, sizeof(ctx->block_flash));
    if (pcsc_rx[0] != 0x90) {
        return -1;
    }

    return 0;
}

static int fwupd_exit(const context_t *ctx) {
    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len;

    pcsc_rx_len = sizeof(pcsc_rx);
    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, ctx->exit, sizeof(ctx->exit));
    if (pcsc_rx[0] != 0x90) {
        return -1;
    }

    return 0;
}

static int get_version() {
    if (fwupd_init_seq() < 0) {
        return -1;
    }

    uint8_t pcsc_rx[EUICC_INTERFACE_BUFSZ];
    uint32_t pcsc_rx_len = sizeof(pcsc_rx);

    memset(pcsc_rx, 0, sizeof(pcsc_rx));
    const int ret = pcsc_transmit_lowlevel(pcsc_rx, &pcsc_rx_len, (const uint8_t *) APDU_GET_VERSION, sizeof(APDU_GET_VERSION) - 1);
    if (ret < 0 || pcsc_rx_len != 8 + 2) {
        return -1;
    }
    printlnf("%s", (const char *) pcsc_rx);

    card_poweroff();
    return 0;
}

static int get_atr(unsigned char **pbAttr, DWORD *dwAttrLen) {
    if (card_poweron() < 0) {
        return -1;
    }
    int ret = SCardGetAttrib(pcsc_hCard, SCARD_ATTR_ATR_STRING, NULL, dwAttrLen);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardGetAttrib()", ret);
        return -1;
    }
    *pbAttr = malloc(*dwAttrLen);
    ret = SCardGetAttrib(pcsc_hCard, SCARD_ATTR_ATR_STRING, *pbAttr, dwAttrLen);
    if (ret != SCARD_S_SUCCESS) {
        pcsc_error("SCardGetAttrib()", ret);
        return -1;
    }
    card_poweroff();
    return 0;
}

static int is_dfu_mode() {
    unsigned char *atr = NULL;
    DWORD atr_len = 0;
    if (get_atr(&atr, &atr_len) < 0) {
        printlnf("GET_ATR_FAILED");
        return -1;
    }
    int ret;
    if (memmem(atr, atr_len, ATR_FPR_FWUPD, sizeof(ATR_FPR_FWUPD) - 1) != NULL) {
        printlnf("FWUPD_MODE");
        ret = 0;
    } else if (memmem(atr, atr_len, ATR_FPR_NOT_FWUPD, sizeof(ATR_FPR_NOT_FWUPD) - 1) != NULL) {
        printlnf("NOT_FWUPD_MODE");
        ret = 1;
    } else {
        printlnf("NOT_ESTKME_PRODUCT");
        ret = -1;
    }
    return ret;
}

static int fwupd(const context_t *ctx, FILE *fw) {
    printf("init fwupd ");
    fflush(stdout);
    if (fwupd_init_seq() < 0) {
        printlnf("failed");
        return -1;
    }
    printlnf("done");

    printf("unlocking ");
    fflush(stdout);
    if (fwupd_unlock_part(ctx) < 0) {
        printlnf("failed");
        return -1;
    }
    printlnf("done");

    const uint8_t mss = get_mss();
    printlnf("mss %d", mss);

    for (int i = 0; !feof(fw); i++) {
        uint8_t block[FWUPD_BLOCK_SIZE];

        switch (fread(block, 1, FWUPD_BLOCK_SIZE, fw)) {
        case FWUPD_BLOCK_SIZE:
            break;
        case 0:
            continue;
        default:
            return -1;
        }

        printf("program block %d ", i);
        fflush(stdout);

        for (int j = 0;; j++) {
            if (j >= 5) {
                fprintlnf(stderr, "send failed");
                return -1;
            }

            if (fwupd_block_write(ctx, block, mss)) {
                continue;
            }

            if (fwupd_block_verify(ctx, block, mss)) {
                continue;
            }

            break;
        }

        if (fwupd_block_flash(ctx)) {
            fprintlnf(stderr, "flash failed");
            return -1;
        }

        printlnf("done");
    }

    if (fwupd_exit(ctx) < 0) {
        fprintlnf(stderr, "fwupd exit failed");
        return -1;
    }

    sleep(5);
    card_poweroff();
    printlnf("fwupd done");
    return 0;
}

int main(const int argc, char *argv[]) {
    if (argc < 2) {
        printlnf("ver, v1-dfu, v2-dfu");
        return -1;
    }

    if (strcmp(argv[1], "ver") == 0) {
        return get_version();
    }

    if (strcmp(argv[1], "is-dfu-mode") == 0) {
        return is_dfu_mode();
    }

    const char *path = getenv("FWUPD_PATH");
    if (argc > 2) {
        path = argv[2];
    } else if (path == NULL) {
        path = "estkme.app.bin";
    }

    FILE *fw = fopen(path, "rb");
    if (fw == NULL) {
        printlnf("%s fopen() failed", path);
        return -1;
    }

    if (strcmp(argv[1], "v1-dfu") == 0) {
        return fwupd(&context_v1, fw);
    }

    if (strcmp(argv[1], "v2-dfu") == 0) {
        return fwupd(&context_v2, fw);
    }

    return -1;
}
